# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = dio.cpp

cdef extern from "../../lib/world/dio.h" nogil:
    cdef cppclass DioOption:
        double f0_floor
        double f0_ceil
        double channels_in_octave
        double frame_period  # msec
        int speed  # (1, 2, ..., 12)
        double allowed_range  # // Threshold used for fixing the F0 contour.
    
    void Dio(double *x, int x_length, int fs, const DioOption option, double *time_axis, double *f0)
    void InitializeDioOption(DioOption *option)
    int GetSamplesForDIO(int fs, int x_length, double frame_period)



# -*- coding: utf-8 -*-
import numpy as np
try:
    from numba import jit
except:
    from pysas.decorators import do_nothing as jit

@jit
def gen_frame(cur_f0, prev_f0, frame, samplingrate, gen_samples, gauss):
    
    if cur_f0 == 0.0 or prev_f0 == 0.0:
        if gauss:
            return np.random.normal(0.0, 1.0, frame), gen_samples
        else:
            return np.random.random(frame), gen_samples
    ret = np.zeros(frame)

    ncur = float(samplingrate) / cur_f0
    nprev = float(samplingrate) / prev_f0
    slope = (nprev - ncur) / float(frame)
    for i in range(frame):
        f0 = ncur + slope * i
        if gen_samples > int(f0):
            ret[i] = f0
            gen_samples -= int(f0)
        gen_samples += 1
    return np.sqrt(ret), gen_samples

@jit
def gen_pulse(f0, frame, samplingrate, gauss):
    ret = np.zeros(f0.shape[0] * frame)
    prev_f0 = f0[0]
    gen_samples = 0
    for i in range(f0.size):
        cur_f0 = f0[i]
        pluses, gen_samples = gen_frame(cur_f0, prev_f0, frame, samplingrate, gen_samples, gauss)
        prev_f0 = cur_f0
        start = frame * i
        ret[start:start+frame] = pluses
    return ret
    



# -*- coding: utf-8 -*-
import numpy as np

padeCoef4 = np.array([1.0, 4.999273e-1, 1.067005e-1, 1.170221e-2, 5.656279e-4])
padeCoef5 = np.array([1.0, 4.999391e-1, 1.107098e-1, 1.369984e-2, 9.564853e-4, 3.041721e-5])


class Filter(object):
    def __init__(self, order, alpha):
        self.order = order
        self.alpha = alpha
        self.delay = np.zeros(order + 1)

    def filter(self, x, coefficients):
        result = 0.0
        self.delay[0] = x
        self.delay[1] = (1.0 - self.alpha ** 2) * self.delay[0] + self.alpha * self.delay[1]
        for i in range(2, coefficients.size):
            self.delay[i] = self.delay[i] + self.alpha * (self.delay[i + 1] - self.delay[i - 1])
            result += self.delay[i] * coefficients[i]
        if coefficients.shape[0] == 2:
            result += self.delay[1] * coefficients[1]
        for i in range(-(self.delay.size - 1), -1):
            i = -i
            self.delay[i] = self.delay[i - 1]
        return result


class CascadeFilter(object):
    def __init__(self, order, alpha, pade_order):
        self.pade_order = pade_order
        self.filter_num = pade_order + 1
        self.order = order
        self.alpha = alpha
        self.filters = []
        self.delay = np.zeros(self.filter_num, dtype=np.float64)
        for i in range(self.filter_num):
            self.filters.append(Filter(order, alpha))

        if pade_order == 4:
            self.pade_coefficients = padeCoef4
        else:
            self.pade_coefficients = padeCoef5
            
    def filter(self, x, coefficients):
        result = 0.0
        feedback = 0.0
        for i in range(-(len(self.pade_coefficients) - 1), 0):
            i = -i
            self.delay[i] = self.filters[i].filter(self.delay[i - 1], coefficients)
            val = self.delay[i] * self.pade_coefficients[i]
            if i%2 == 1:
                feedback += val
            else:
                feedback -= val
            result += val
        self.delay[0] = feedback + x
        result += self.delay[0]
        return result

class MLSAFilter(object):
    def __init__(self, order, alpha, pade_order):

        assert pade_order == 4 or pade_order == 5, "order of pade must be 4 or 5."
        self.f1 = CascadeFilter(2, alpha, pade_order)
        self.f2 = CascadeFilter(order + 1, alpha, pade_order)

    def filter(self, x, coefficients):
        coef = np.array([0, coefficients[1]])
        return self.f2.filter(self.f1.filter(x, coef),coefficients)
        



# -*- coding: utf-8 -*-
from  unittest import TestCase
from nose.tools import eq_
import numpy as np

from pysas import waveread, World
from pysas.mcep import estimate_alpha, spec2mcep_from_matrix, mcep2coef
from pysas.synthesis.mlsa import MLSAFilter
from pysas.synthesis import Synthesis
from pysas.excite import ExcitePulse

class SynthesisTest(TestCase):
    def setUp(self):
        signal, samplingrate, _ = waveread("test/cmu_arctic/arctic_a0001.wav")
        self.world = World(samplingrate)
        self.alpha = estimate_alpha(samplingrate)
        self.samplingrate = samplingrate
        self.signal = signal
        self.f0, self.spec_mat, _ = self.world.analyze(signal)
        self.ep = ExcitePulse(16000, 80, False)
        self.order = 24
        
    def test_synthesis_filter(self):
        excite = self.ep.gen(self.f0)
        mcep_mat = spec2mcep_from_matrix(self.spec_mat, self.order, self.alpha)
        coef_mat = []
        for i in range(mcep_mat.shape[0]):
            coef_mat.append(mcep2coef(mcep_mat[i], 0.41))
        coef_mat = np.array(coef_mat)
        mlsa = MLSAFilter(self.order, self.alpha, 5)
        syn = Synthesis(80, mlsa)
        synth = syn.synthesis(excite, coef_mat)



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = mcep.cpp

import numpy as np
cimport numpy as np
cimport cython
from cython.view cimport array as cyarray
from libc.string cimport memcpy, memset
from libc.stdlib cimport malloc, free
from libc.math cimport log, M_PI, sin, cos, atan
from libc.float cimport DBL_MAX


@cython.boundscheck(False)
@cython.wraparound(False)
def spec2mcep_from_matrix(np.ndarray[np.float64_t, ndim=2, mode="c"] spmat, int order, double alpha):
    """
    calucurate mel-cepstrum from spectrogram matrix
    """
    cdef int fftsize = (spmat.shape[1] - 1) * 2
    cdef np.ndarray[np.float64_t, ndim=2, mode="c"] logspmat = np.log(spmat)
    cdef np.ndarray[np.float64_t, ndim=2, mode="c"] c_mat = np.real(np.fft.irfft(logspmat, fftsize))

    cdef np.ndarray[np.float64_t, ndim=2, mode="c"] wc_mat = np.zeros(
        (c_mat.shape[0], order + 1),dtype=np.double)
    
    cdef int i
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] c_vec, wc_vec, prev
    prev = np.zeros(wc_mat.shape[1], dtype=np.float64)
    for i in range(c_mat.shape[0]):
        # ToDo: change ndarray to typed memoryview
        wc_vec = wc_mat[i]
        c_vec = c_mat[i]
        c_vec[0] /= 2.0
        freqt(<double *>c_vec.data, <double *>wc_vec.data, <double *>prev.data, c_vec.size, alpha, order)
    return wc_mat


@cython.boundscheck(False)
@cython.wraparound(False)
def spec2mcep(np.ndarray[np.float64_t, ndim=1, mode="c"] sp, int order, double alpha):
    """
    calucurate mel-cepstrum from spectrogram vector
    """
    cdef int fftsize = (sp.size - 1) * 2
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] logsp = np.log(sp)
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] c = np.real(np.fft.irfft(logsp, fftsize))
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] wc = np.zeros(order + 1, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] prev = np.zeros_like(wc)
    c[0] /= 2.0
    freqt(<double *>c.data, <double *>wc.data, <double *>prev.data, c.size, alpha, order)
    return wc


@cython.boundscheck(False)
@cython.wraparound(False)
def mcep2spec(np.ndarray[np.float64_t, ndim=1, mode="c"] mcep, double alpha, int fftsize):
    cdef int fsize = fftsize >> 1
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] c = np.zeros(fsize + 1, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] prev = np.zeros_like(c)
    freqt(<double *>mcep.data, <double *>c.data, <double *>prev.data, mcep.size, -alpha, fsize)
    c[0] *= 2.0
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] ret = c.copy()
    ret.resize(fftsize, refcheck=False)
    ret = ret
    cdef int i
    for i in range(fsize):
        ret[fftsize - i - 1] = ret[i+1]
    return np.exp(np.real(np.fft.rfft(ret)))


@cython.boundscheck(False)
@cython.wraparound(False)
def mcep2spec_from_matrix(np.ndarray[np.float64_t, ndim=2, mode="c"] mcepmat, double alpha, int fftsize):
    cdef int fsize = fftsize >> 1
    cdef np.ndarray[np.float64_t, ndim=2, mode="c"] c = np.zeros((mcepmat.shape[0], fsize + 1), dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] prev = np.zeros(fsize + 1, dtype=np.float64)
    cdef np.ndarray[np.float64_t, ndim=2, mode="c"] ret = np.zeros((mcepmat.shape[0], fftsize), dtype=np.float64)
    cdef int i
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] c_vec, mcep_vec, ret_vec
    for i in range(mcepmat.shape[0]):
        # ToDo: change ndarray to typed memoryview
        c_vec = c[i]
        mcep_vec = mcepmat[i]
        ret_vec = ret[i]
        freqt(<double *>mcep_vec.data, <double *>c_vec.data, <double *>prev.data, mcep_vec.size, -alpha, fsize)
        c_vec[0] *= 2.0
        np.copyto(ret_vec[:fsize + 1], c_vec)
        for i in range(fsize):
            ret_vec[fftsize - i - 1] = ret_vec[i+1]
    return np.exp(np.real(np.fft.rfft(ret)))

@cython.boundscheck(False)
@cython.wraparound(False)
def mcep2coef(np.ndarray[np.float64_t, ndim=1, mode="c"] mcep, double alpha):
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] coef = mcep.copy()
    cdef int i
    for i from mcep.size - 2 >= i >= 0:
        coef[i] = coef[i] - alpha * coef[i+1]
    return coef

@cython.boundscheck(False)
@cython.wraparound(False)
def coef2mcep(np.ndarray[np.float64_t, ndim=1, mode="c"] coef, double alpha):
    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] mcep = np.copy(coef)
    cdef size = coef.size
    cdef double d = mcep[size - 1]
    cdef double o = 0
    cdef int i
    for i from size - 2 >= i >= 0:
        o = mcep[i] + alpha * d
        d = mcep[i]
        mcep[i] = o
    return mcep


@cython.boundscheck(False)
@cython.wraparound(False)
cdef void freqt(const double *c, double *wc, double *prev, int c_size, double alpha, int order) nogil:
    cdef int i, j
    for i in range(-(c_size - 1), 1):  # -(c_size - 1) 〜 0
        memcpy(prev, wc, sizeof(double)*(order + 1))
        if order >= 0:
            wc[0] = c[-i] + alpha * prev[0]
        if order >= 1:
            wc[1] = (1.0 - alpha ** 2) * prev[0] + alpha * prev[1]
        for j in range(2, order + 1):  # 2 ~ order -> length(wc) -> order + 1
            wc[j] = prev[j - 1] + alpha * (prev[j] - wc[j - 1])


@cython.boundscheck(False)
@cython.wraparound(False)
def estimate_alpha(int sampfreq, double start=0, double end=1.0, double step=0.001, int size=1000):
    """
    estimate alpha parameter for mel-cepstrum from sampling rate.
    original imprementation https://bitbucket.org/happyalu/mcep_alpha_calc
    """
    cdef double *melscale_vector = create_melscale_vector(sampfreq, size)
    cdef double *warping_vector = <double *>malloc(size * sizeof(double))
    cdef double i, 
    cdef double min_dist, best_alpha, dist
    best_alpha = 0.0
    min_dist = DBL_MAX
    for a from start <= a <= end by step:
        calc_warping_vector(warping_vector, a, size)
        dist = rms_distance_like(melscale_vector, warping_vector, size)
        if dist < min_dist:
            min_dist = dist
            best_alpha = a
    free(melscale_vector)
    free(warping_vector)
    return best_alpha
    

cdef double* create_melscale_vector(int sampfreq, int size) nogil:
    cdef double *vec = <double *>malloc(size * sizeof(double))
    cdef double step = (sampfreq / 2.0) / size
    cdef int i
    for i in range(size):
        vec[i] = (1000.0 / log(2)) * log(1.0+((step * i)/1000.0))
    cdef double last = vec[size - 1]
    for i in range(size):
        vec[i] /= last
    return vec


cdef void calc_warping_vector(double *vec, double alpha, int size) nogil:
    cdef double step = M_PI / size
    cdef int i
    cdef double omega, warpfreq, num, den
    for i in range(size):
        omega = step * i
        num = (1 - alpha * alpha) * sin(omega)
        den = (1 + alpha * alpha) * cos(omega) - 2 * alpha
        warpfreq = atan(num / den)
        if warpfreq < 0:
            warpfreq += M_PI
        vec[i] = warpfreq
    cdef double last = vec[size - 1]
    for i in range(size):
        vec[i] /= last

cdef double rms_distance_like(double *a, double *b, int size) nogil:
    cdef int i
    cdef double s = 0.0
    for i in range(size):
        s += (a[i] - b[i]) ** 2
    return s



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = synthesis.cpp

cdef extern from "../../lib/world/synthesis.h" nogil:
    void Synthesis(double *f0, int f0_length, double **spectrogram, double **aperiodicity,
                   int fft_size, double frame_period, int fs,
                   int y_length, double *y)



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = stonemask.cpp

cdef extern from "../../lib/world/stonemask.h" nogil:
    void StoneMask(double *x, int x_length, int fs, double *time_axis, double *f0,
                   int f0_length, double *refined_f0);



# -*- coding: utf-8 -*-
from  unittest import TestCase
from nose.tools import eq_
import numpy as np

from pysas import waveread
from pysas.mcep import estimate_alpha, spec2mcep, mcep2spec, mcep2coef, coef2mcep
from pysas.mcep import mcep2spec_from_matrix, spec2mcep_from_matrix
from pysas.synthesis_filter.mlsa import MLSAFilter
from pysas import World


class MecpTest(TestCase):
    def setUp(self):
        signal, samplingrate, _ = waveread("test/cmu_arctic/arctic_a0001.wav")
        self.signal = signal
        self.samplingrate = samplingrate
        start = 80*200
        self.windowsize = 1024
        signal = signal[start:start+self.windowsize] * np.blackman(self.windowsize)
        self.pspec = np.absolute(np.fft.fft(signal) ** 2)[:(self.windowsize>>1) + 1]
        self.alpha = 0.41

    def test_spec2mcep(self):
        mcep = spec2mcep(self.pspec, 20, self.alpha)
        spec = mcep2spec(mcep, self.alpha, self.windowsize)
        sqerr = np.sqrt((np.log(self.pspec) - np.log(spec)) ** 2).sum() / self.pspec.size
        assert sqerr < 1.2, sqerr
        
    def test_mcep2coef(self):
        mcep = spec2mcep(self.pspec, 20, self.alpha)
        coef = mcep2coef(mcep, self.alpha)
        mcep2 = coef2mcep(coef, self.alpha)
        sqerr = np.sqrt((mcep - mcep2) ** 2).sum()
        assert sqerr < 0.0001, sqerr

    def test_from_matrix(self):
        world = World(self.samplingrate)
        _, spec, _ = world.analyze(self.signal)
        mcepmat = spec2mcep_from_matrix(spec, 20, self.alpha)
        mcep = spec2mcep(spec[300], 20, self.alpha)
        assert (mcep == mcepmat[300]).all()
        specmat = mcep2spec_from_matrix(mcepmat, self.alpha, self.windowsize)
        spec = mcep2spec(mcep, self.alpha, self.windowsize)
        assert (spec == specmat[300]).all()
        
class EstimateAlphaTest(TestCase):
    def _callFUT(self, sampfreq):
        return estimate_alpha(sampfreq)
    def test_8k(self):
        eq_(round(self._callFUT(8000), 3), 0.312)
    def test_16k(self):
        eq_(round(self._callFUT(16000), 3), 0.41)
    def test_44k(self):
        eq_(round(self._callFUT(44100), 3), 0.544)




# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = mlsa.cpp

# porting from https://gist.github.com/r9y9/7735120

import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free
from cpython cimport PyObject, Py_INCREF, Py_DECREF


cdef double[:] pade_coef4 = np.array([1.0, 4.999273e-1, 1.067005e-1, 1.170221e-2, 5.656279e-4])
cdef double[:] pade_coef5 = np.array([1.0, 4.999391e-1, 1.107098e-1, 1.369984e-2, 9.564853e-4, 3.041721e-5])


cdef class Filter:
    cdef double alpha
    cdef int order
    cdef double[:] delay
    
    def __cinit__(self, int order, double alpha):
        self.order = order
        self.alpha = alpha
        self.delay = np.zeros(order + 1, dtype=np.float64)

    cdef filter(self, double x, np.ndarray[np.float64_t, ndim=1, mode="c"] coefficients):
        cdef double result = 0.0
        self.delay[0] = x
        self.delay[1] = (1.0 - self.alpha ** 2) * self.delay[0] + self.alpha * self.delay[1]
        cdef int i
        for i in range(2, coefficients.size):
            self.delay[i] = self.delay[i] + self.alpha * (self.delay[i + 1] - self.delay[i - 1])
            result += self.delay[i] * coefficients[i]
        if coefficients.shape[0] == 2:
            result += self.delay[1] * coefficients[1]
        for i in range(-(self.delay.size - 1), -1):
            i = -i
            self.delay[i] = self.delay[i - 1]
        return result

cdef class CascadeFilter:
    cdef int pade_order, order, filter_num
    cdef double alpha
    cdef double[:] delay
    cdef double[:] pade_coefficients
    cdef PyObject **filters
    def __cinit__(self, int order, double alpha, int pade_order):
        self.pade_order = pade_order
        self.filter_num = pade_order + 1
        self.order = order
        self.alpha = alpha
        self.delay = np.zeros(self.filter_num, dtype=np.float64)
        self.filters = <PyObject **>malloc(sizeof(PyObject *) * self.filter_num)
        for i in range(self.filter_num):
            filt = Filter(order, alpha)
            Py_INCREF(filt)
            self.filters[i] = <PyObject *>filt
        if pade_order == 4:
            self.pade_coefficients = pade_coef4
        else:
            self.pade_coefficients = pade_coef5

    def __dealloc__(self):
        cdef int i
        for i in range(self.filter_num):
            Py_DECREF(<object>self.filters[i])
        free(self.filters)

    cdef filter(self, double x, np.ndarray[np.float64_t, ndim=1, mode="c"] coefficients):
        cdef double result = 0.0
        cdef double feedback = 0.0
        cdef int i
        for i in range(-(len(self.pade_coefficients) - 1), 0):
            i = -i
            self.delay[i] = (<Filter>self.filters[i]).filter(self.delay[i - 1], coefficients)
            val = self.delay[i] * self.pade_coefficients[i]
            if i%2 == 1:
                feedback += val
            else:
                feedback -= val
            result += val
        self.delay[0] = feedback + x
        result += self.delay[0]
        return result

cdef class MLSAFilter(object):
    cdef int pade_order, order
    cdef double alpha
    cdef CascadeFilter f1
    cdef CascadeFilter f2
    
    def __init__(self, int order, double alpha, int pade_order):
        assert pade_order == 4 or pade_order == 5, "order of pade must be 4 or 5."
        self.f1 = CascadeFilter(2, alpha, pade_order)
        self.f2 = CascadeFilter(order + 1, alpha, pade_order)

    def filter(self, x, coefficients):
        coef = np.array([0, coefficients[1]])
        return self.f2.filter(self.f1.filter(x, coef),coefficients)
        



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = mlsa.cpp

import numpy as np
cimport numpy as np


cdef class Synthesis(object):
    cdef int frameperiod
    cdef object filter
    
    def __init__(self, int frameperiod, object filter):
        self.frameperiod = frameperiod
        self.filter = filter

    def synthesis_frame(self, np.ndarray[np.float64_t, ndim=1, mode="c"] cur_coef,
                        np.ndarray[np.float64_t, ndim=1, mode="c"] prev_coef,
                        np.ndarray[np.float64_t, ndim=1, mode="c"] pulse):
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] ret = np.zeros_like(pulse)
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] slope = (cur_coef - prev_coef) / self.frameperiod
        cdef int i
        cdef double scaled
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] coef
        for i in range(self.frameperiod):
            coef = prev_coef + slope * i
            scaled = pulse[i] * np.exp(coef[0])
            ret[i] = self.filter.filter(scaled, coef)
        return ret
        
    def synthesis(self, np.ndarray[np.float64_t, ndim=1, mode="c"] pulse,
                  np.ndarray[np.float64_t, ndim=2, mode="c"] coef_mat):
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] ret = np.zeros_like(pulse)
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] prev_coef = np.zeros_like(coef_mat[0])
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] cur_coef
        cdef int i, offset
        for i in range(coef_mat.shape[0]):
            offset = self.frameperiod * i
            cur_coef = coef_mat[i]
            ret[offset:offset+self.frameperiod] = self.synthesis_frame(cur_coef, prev_coef, pulse[offset:offset+self.frameperiod])
            prev_coef = cur_coef
        return ret



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = world.cpp
import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free
from libc.math cimport log, exp

from libworld cimport dio
from libworld cimport d4c
from libworld cimport stonemask
from libworld cimport cheaptrick
from libworld cimport matlabfunctions
from libworld cimport synthesis

    
cdef class World:
    """
    cython Wrapper For World
    """
    cdef int samplingrate, fft_size, envelope_size
    cdef double frameperiod
    
    def __init__(self, int samplingrate, double frameperiod=5.0):
        """
        @param samplingrate sampling rate of signal(integaer)
        @param frameperiod frame period(msec, default=5.0msec)
        """
        self.frameperiod = frameperiod  # ms
        self.samplingrate = samplingrate
        self.fft_size = cheaptrick.GetFFTSizeForCheapTrick(self.samplingrate)
        self.envelope_size = self.fft_size / 2 + 1

    def fftsize(self):
        return self.fft_size

    def envelopesize(self):
        return self.envelope_size
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def analyze(self, np.ndarray[np.float64_t, ndim=1, mode="c"] signal):
        """
        @return f0, spectral envelope, aperiodicity
        """
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] f0, time_axis
        f0, time_axis = self.estimate_f0(signal)
        
        cdef np.ndarray[np.float64_t, ndim=2, mode="c"] spectrogram, aperiodicity
        spectrogram = np.zeros((f0.size, self.envelope_size), dtype=np.float64)
        aperiodicity = np.zeros((f0.size, self.envelope_size), dtype=np.float64)

        cdef double **c_spectrogram = <double **>malloc(f0.size * sizeof(double *))
        cdef double **c_aperiodicity = <double **>malloc(f0.size * sizeof(double *))

        cdef int i
        # copy pointer to c array
        for i in range(f0.size):
            c_spectrogram[i] = &(<double *>spectrogram.data)[i * spectrogram.shape[1]]
            c_aperiodicity[i] = &(<double *>aperiodicity.data)[i * aperiodicity.shape[1]]

        cheaptrick.CheapTrick(<double *>signal.data, signal.size, self.samplingrate,
                              <double *>time_axis.data, <double *>f0.data,
                              f0.size, c_spectrogram)

        d4c.D4C(<double *>signal.data, signal.size, self.samplingrate,
                <double *>time_axis.data, <double *>f0.data,
                f0.size, self.fft_size, c_aperiodicity)

        free(c_spectrogram)
        free(c_aperiodicity)
        return f0, spectrogram, aperiodicity

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def spectral_envelope(self, np.ndarray[np.float64_t, ndim=1, mode="c"] signal):
        """
        estimate F0 and spectral envelope from speech signal
        @signal speech signal
        @return f0, spectral envelope
        """
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] f0, time_axis
        f0, time_axis = self.estimate_f0(signal)
        
        cdef np.ndarray[np.float64_t, ndim=2, mode="c"] spectrogram
        spectrogram = np.zeros((f0.size, self.envelope_size), dtype=np.float64)

        cdef double **c_spectrogram = <double **>malloc(f0.size * sizeof(double *))

        cdef int i
        # copy pointer to c array
        for i in range(f0.size):
            c_spectrogram[i] = &(<double *>spectrogram.data)[i * spectrogram.shape[1]]

        cheaptrick.CheapTrick(<double *>signal.data, signal.size, self.samplingrate,
                              <double *>time_axis.data, <double *>f0.data,
                              f0.size, c_spectrogram)

        free(c_spectrogram)
        return f0, spectrogram
    
    def f0_scaling(self, np.ndarray[np.float64_t, ndim=1, mode="c"] f0, double shift):
        if shift <= 0:
            return f0
        return f0 * shift
    
    @cython.boundscheck(False)
    @cython.wraparound(False)
    def spectral_stretching(self, np.ndarray[np.float64_t, ndim=2, mode="c"] spectrogram,
                            double ratio):
        if ratio <= 0:
            return spectrogram
        cdef int dim = spectrogram.shape[1]
        assert dim == self.envelope_size, "dimension of spectrogram should be same as self.envelope_size"
        cdef double *freq_axis1 = <double *>malloc(dim * sizeof(double))
        cdef double *freq_axis2 = <double *>malloc(dim * sizeof(double))
        cdef double *spectrum1 = <double *>malloc(dim * sizeof(double))
        cdef double *spectrum2 = <double *>malloc(dim * sizeof(double))

        cdef int i, j, k
        k = <int>(self.fft_size / 2.0 * ratio)
        for i in range(dim):
            freq_axis1[i] = ratio * i /self.fft_size * self.samplingrate
            freq_axis2[i] = <double>i / self.fft_size * self.samplingrate

        for i in range(spectrogram.size):
            for j in range(dim):
                spectrum1[j] = log(spectrogram[i, j])
            matlabfunctions.interp1(freq_axis1, spectrum1, dim,
                                    freq_axis2, dim, spectrum2);
            for j in range(dim):
                spectrogram[i, j] = exp(spectrum2[j])
            if (ratio < 1.0 and k > 0):
                for j in range(k, dim):
                    spectrogram[i, j] = spectrogram[i, k - 1]
        free(freq_axis1)
        free(freq_axis2)
        free(spectrum1)
        free(spectrum2)
        return spectrogram

    def synthesis(self,
                  np.ndarray[np.float64_t, ndim=1, mode="c"] f0,
                  np.ndarray[np.float64_t, ndim=2, mode="c"] spectrogram,
                  np.ndarray[np.float64_t, ndim=2, mode="c"] aperiodicity):
        assert f0.size == spectrogram.shape[0] == aperiodicity.shape[0], "all arguments should be same sample size"
        assert spectrogram.shape[1] == aperiodicity.shape[1], "spectrogram and aperiodicity should be same size"
                
        cdef double **c_spectrogram = <double **>malloc(f0.size * sizeof(double *))
        cdef double **c_aperiodicity = <double **>malloc(f0.size * sizeof(double *))

        cdef int result_length = <int>((f0.size - 1) * self.frameperiod / 1000.0 * self.samplingrate) + 1
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] result = np.zeros(result_length, dtype=np.float64)

        cdef int i
        # copy to c array
        for i in range(spectrogram.shape[0]):
            c_spectrogram[i] = &(<double *>spectrogram.data)[i * spectrogram.shape[1]]
            c_aperiodicity[i] = &(<double *>aperiodicity.data)[i * aperiodicity.shape[1]]
        
        synthesis.Synthesis(<double *>f0.data, f0.size, c_spectrogram, c_aperiodicity,
                            self.fft_size, self.frameperiod, self.samplingrate,
                            result_length, <double *>result.data)

        free(c_spectrogram)
        free(c_aperiodicity)
        return result
    
    cdef estimate_f0(self, np.ndarray[np.float64_t, ndim=1, mode="c"] signal):
        """
        estimate F0 from speech signal
        @signal speech signal
        @return estimated f0 series
        """
        f0_length = dio.GetSamplesForDIO(self.samplingrate, signal.size, self.frameperiod)
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] f0, refined_f0, time_axis
        f0 = np.zeros(f0_length, dtype=np.float64)
        refined_f0 = np.zeros(f0_length, dtype=np.float64)
        time_axis = np.zeros(f0_length, dtype=np.float64)

        cdef dio.DioOption option
        dio.InitializeDioOption(&option)
        
        # modify parameter
        option.frame_period = self.frameperiod
        option.speed = 1
        option.f0_floor = 71.0
        option.allowed_range = 0.1
  
        dio.Dio(<double *> signal.data, signal.size, self.samplingrate,
                option, <double *> time_axis.data, <double *> f0.data)
        stonemask.StoneMask(<double *> signal.data, signal.size, self.samplingrate,
                            <double *> time_axis.data, <double *> f0.data,
                            f0_length, <double *> refined_f0.data)
        return refined_f0, time_axis
    



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = cheaptrick.cpp

cdef extern from "../../lib/world/cheaptrick.h" nogil:
    void CheapTrick(double *x, int x_length, int fs, double *time_axis, double *f0,
                    int f0_length, double **spectrogram)
    int GetFFTSizeForCheapTrick(int fs)



# -*- coding: utf-8 -*-
from  unittest import TestCase
from nose.tools import eq_
import numpy as np

from pysas import World, waveread


class WorldTest(TestCase):
    def setUp(self):
        signal, samplingrate, _ = waveread("test/cmu_arctic/arctic_a0001.wav")
        self.signal = signal
        self.sampfreq = samplingrate
        self.world = World(samplingrate)

    def _callFUT(self, signal):
        return self.world.analyze(signal)

    def test_analyze(self):
        f0, spec, aperiod = self._callFUT(self.signal)
        eq_(f0.shape[0], spec.shape[0], aperiod.shape[0])
        out = self.world.synthesis(f0, spec, aperiod)
        assert len(out) == len(self.signal), "{}, {}".format(len(out), len(self.signal))
        



# -*- coding: utf-8 -*-
import numpy as np

class Synthesis(object):
    def __init__(self, frameperiod, filter):
        self.frameperiod = frameperiod
        self.filter = filter

    def synthesis_frame(self, cur_coef, prev_coef, pulse):
        ret = np.zeros_like(pulse)
        slope = (cur_coef - prev_coef) / self.frameperiod
        for i in range(self.frameperiod):
            coef = prev_coef + slope * i
            scaled = pulse[i] * np.exp(coef[0])
            ret[i] = self.filter.filter(scaled, coef)
        return ret
        
    def synthesis(self, pulse, coef_mat):
        ret = np.zeros_like(pulse)
        prev_coef = np.zeros_like(coef_mat[0])
        for i in range(coef_mat.shape[0]):
            offset = self.frameperiod * i
            cur_coef = coef_mat[i]
            synthesized = self.synthesis_frame(cur_coef, prev_coef, pulse[offset:offset+self.frameperiod])
            if i > 0:
                ret[offset:offset+self.frameperiod] = synthesized
            prev_coef = cur_coef
        return ret



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = d4c.cpp

cdef extern from "../../lib/world/d4c.h" nogil:
    void D4C(double *x, int x_length, int fs, double *time_axis, double *f0,
             int f0_length, int fft_size, double **aperiodicity)



# -*- coding: utf-8 -*-
from setuptools import setup, Extension, find_packages
from Cython.Distutils import build_ext
import sys
import numpy as np
from glob import glob
sys.path.append('./src')
sys.path.append('./test')

ext_modules = [Extension(
        'pysas.world',
        ["pysas/world.pyx"] + glob("lib/world/*.cpp"),
        include_dirs=['lib/world', np.get_include()],
        extra_compile_args=["-O3"],
        language="c++",
    ),
    Extension(
        'pysas.mcep',
        ["pysas/mcep.pyx"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-O3"],
        language="c++",
    ),
    Extension(
        'pysas.excite',
        ["pysas/excite.pyx"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-O3"],
        language="c++",
    ),
    Extension(
        'pysas.synthesis.mlsa',
        ["pysas/synthesis/mlsa.pyx"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-O3"],
        language="c++",
    )
]
    
setup(
    name="pysas",
    description='Speech Analysis and Synthesis for Python',
    version="0.02",
    long_description=open('README.rst').read(),
    packages=find_packages(),
    install_requires=["numpy","cython", 'nose'],
    ext_modules=ext_modules,
    cmdclass={'build_ext': build_ext},
    author='Shunsuke Aihara',
    author_email="aihara@argmax.jp",
    url='https://github.com/shunsukeaihara/pysas',
    license="MIT License",
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['Nose', 'cython', 'numpy'],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering"
    ])



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = matlabfunctions.cpp

cdef extern from "../../lib/world/matlabfunctions.h" nogil:
    void interp1(double *x, double *y, int x_length, double *xi, int xi_length, double *yi)



# -*- coding: utf-8 -*-
# distutils: language = c++
# distutils: sources = excite.cpp

import numpy as np
cimport numpy as np
cimport cython
from cpython cimport bool


cdef class ExcitePulse:
    cdef int gen_samples, frameperiod, samplingrate
    cdef bool gauss
    def __cinit__(self, int samplingrate, int frameperiod, bool gauss):
        self.gen_samples = 0
        self.samplingrate = samplingrate
        self.frameperiod = frameperiod
        self.gauss = gauss

    cdef np.ndarray[np.float64_t, ndim=1, mode="c"] gen_frame(self, double cur_f0, double prev_f0):
        if cur_f0 == 0.0 or prev_f0 == 0.0:
            if self.gauss:
                return np.random.normal(0.0, 1.0, self.frameperiod)
            else:
                return np.random.random(self.frameperiod)
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] ret = np.zeros(self.frameperiod)
        cdef double cur_hz = <double>self.samplingrate / cur_f0
        cdef double prev_hz = <double>self.samplingrate / prev_f0
        cdef double slope = (cur_hz - prev_hz) / <double>self.frameperiod
        cdef int i
        cdef double f0
        for i in range(self.frameperiod):
            f0 = prev_hz + slope * i
            if self.gen_samples > <int>f0:
                ret[i] = f0
                self.gen_samples -= <int>f0
            self.gen_samples += 1
        return np.sqrt(ret)
        
    def gen(self, np.ndarray[np.float64_t, ndim=1, mode="c"] f0):
        ret = np.zeros(f0.shape[0] * self.frameperiod)
        self.gen_samples = 0
        cdef int i, offset
        cdef double cur_f0, prev_f0
        cdef np.ndarray[np.float64_t, ndim=1, mode="c"] pluses
        prev_f0 = f0[0]
        for i in range(f0.size):
            cur_f0 = f0[i]
            pluses = self.gen_frame(cur_f0, prev_f0)
            prev_f0 = cur_f0
            offset = self.frameperiod * i
            ret[offset:offset+self.frameperiod] = pluses
        return ret



# -*- coding: utf-8 -*-
import wave
import numpy as np

def waveread(path):
    """
    read signals from wavefile, only support monoral and 8 or 16bit data
    @return normalized signals between [-1,1], sampling rate[Hz], bit rate[bits]
    """
    wf = wave.open(path, 'rb')
    channels = wf.getnchannels()
    assert channels == 1, "wave file should be monoral"
    
    bits = wf.getsampwidth() * 8  # bit rate
    assert bits==8 or bits==16, "bit rate should be 8 or 16bits (input {}bit)".format(bits)
    
    fs = wf.getframerate()  # sampling rate(Hz)
    datalength = wf.getnframes()
    raw = wf.readframes(datalength)

    
    if bits == 8:
        signal = np.float64(np.fromstring(raw, dtype=np.uint8) - 127) / 127.0
    elif bits == 16:
        signal = np.float64(np.fromstring(raw, dtype=np.int16)) / 32767.0
    return signal, fs, bits


def wavewrite(signal, samplingrate, bits, path):
    assert bits==8 or bits==16, "bit rate expect 8 or 16bits but recevied {}bit)".format(bits)
    wf=wave.open(path,'wb')
    if bits == 8:
        sampwidth = 1
        s = np.int8((signal + 1.0) * 127).tostring()
    else:
        sampwidth = 2
        s = np.int16(signal * 32767.0).tostring()
        
    params = (1, sampwidth, samplingrate, len(signal), 'NONE', 'not compressed')
    wf.setparams(params)
    wf.writeframes(s)
